﻿using System;
using System.Reflection;
using System.IO;
using Microsoft.SharePoint.Administration;
using System.Runtime.InteropServices;

namespace TechEd09.SharePoint.Diagnostics
{
    public static class TraceLogging
    {
        static object s_lock = new object();
        static bool s_initialized = false;
        static UInt64 s_traceLog;
        static UInt64 s_traceReg;

        public enum TraceSeverity
        {
            Unassigned = 0,
            CriticalEvent = 1,
            WarningEvent = 2,
            InformationEvent = 3,
            Exception = 4,
            Assert = 7,
            Unexpected = 10,
            Monitorable = 15,
            High = 20,
            Medium = 50,
            Verbose = 100,
        };

        public static void WriteTrace(  string exeName,
                                        string productName,
                                        string categoryName,
                                        string message)
        {
            WriteTrace( TraceSeverity.Unassigned,
                        exeName,
                        productName,
                        categoryName,
                        message);
        }

        public static void WriteTrace(  TraceSeverity severity,
                                        string exeName,
                                        string productName,
                                        string categoryName,
                                        string message)
        {
            WriteTrace(null, severity, exeName, productName, categoryName, message);
        }

        public static void WriteTrace(  string tag, 
                                        TraceSeverity severity,
                                        string exeName,
                                        string productName, 
                                        string categoryName, 
                                        string message)
        {
            EnsureInitialized();

            uint t = TagFromString(tag);

            WriteTraceInternal( t,
                                severity,
                                exeName,
                                productName,
                                categoryName,
                                message);

        }

        static void WriteTraceInternal( uint tag,
                                        TraceSeverity severity,
                                        string exeName,
                                        string productName,
                                        string categoryName,
                                        string message)
        {
            const ushort sizeOfWCHAR = 2;
            NativeMethods.ULSTrace ulsTrace = new NativeMethods.ULSTrace();

            // Pretty standard code needed to make things work
            ulsTrace.Header.Size = (ushort)Marshal.SizeOf(typeof(NativeMethods.ULSTrace));
            ulsTrace.Header.Flags = NativeMethods.WNODE_FLAG_TRACED_GUID;
            ulsTrace.ULSHeader.dwVersion = NativeMethods.TRACE_VERSION_CURRENT;
            ulsTrace.ULSHeader.dwFlags = NativeMethods.TraceFlags.TRACE_FLAG_ID_AS_ASCII;
            ulsTrace.ULSHeader.Size = (ushort)Marshal.SizeOf(typeof(NativeMethods.ULSTraceHeader));

            // Variables communicated to SPTrace
            ulsTrace.ULSHeader.Id = tag;
            ulsTrace.Header.Class.Level = (byte)severity;
            ulsTrace.ULSHeader.wzExeName = exeName;
            ulsTrace.ULSHeader.wzProduct = productName;
            ulsTrace.ULSHeader.wzCategory = categoryName;
            ulsTrace.ULSHeader.wzMessage = message;
            ulsTrace.ULSHeader.correlationID = Guid.Empty;

            // Pptionally, to improve performance by reducing the amount of data copied around,
            // the Size parameters can be reduced by the amount of unused buffer in the Message
            if (message.Length < 800)
            {
                ushort unusedBuffer = (ushort)((800 - (message.Length + 1)) * sizeOfWCHAR);
                ulsTrace.Header.Size -= unusedBuffer;
                ulsTrace.ULSHeader.Size -= unusedBuffer;
            }

            if (s_traceLog != 0)
                NativeMethods.TraceEvent(s_traceLog, ref ulsTrace);
        }

        static void EnsureInitialized()
        {
            if (!s_initialized)
            {
                lock (s_lock)
                {
                    if (!s_initialized)
                    {
                        InitializeTraceLogging();
                    }
                }
            }
        }

        static void InitializeTraceLogging()
        {
            RegisterTraceProvider();

            AppDomain.CurrentDomain.DomainUnload += new EventHandler(CurrentDomain_DomainUnload);
            AppDomain.CurrentDomain.ProcessExit += new EventHandler(CurrentDomain_ProcessExit);
            s_initialized = true;
        }

        static void StopTraceLogging()
        {
            if (s_initialized)
            {
                lock (s_lock)
                {
                    if (s_initialized)
                    {
                        UnregisterTraceProvider();
                        s_initialized = false;
                    }
                }
            }
        }

        static void CurrentDomain_ProcessExit(object sender, EventArgs e)
        {
            StopTraceLogging();
        }

        static void CurrentDomain_DomainUnload(object sender, EventArgs e)
        {
            StopTraceLogging();
        }

        static unsafe void RegisterTraceProvider()
        {
            SPFarm farm = SPFarm.Local;
            Guid traceGuid = farm.TraceSessionGuid;
            uint result = NativeMethods.RegisterTraceGuids( ControlCallback, 
                                                            null, 
                                                            ref traceGuid, 
                                                            0, 
                                                            IntPtr.Zero, 
                                                            null, 
                                                            null,
                                                            out s_traceReg);

            System.Diagnostics.Debug.Assert(result == NativeMethods.ERROR_SUCCESS);
        }

        static void UnregisterTraceProvider()
        {
            uint result = NativeMethods.UnregisterTraceGuids(s_traceReg);
            System.Diagnostics.Debug.Assert(result == NativeMethods.ERROR_SUCCESS);
        }

        static unsafe uint ControlCallback( NativeMethods.WMIDPREQUESTCODE RequestCode, 
                                            IntPtr Context, 
                                            uint* InOutBufferSize, 
                                            IntPtr Buffer)
        {
            uint Status;
            switch (RequestCode)
            {
                case NativeMethods.WMIDPREQUESTCODE.WMI_ENABLE_EVENTS:
                    s_traceLog = NativeMethods.GetTraceLoggerHandle(Buffer);
                    Status = NativeMethods.ERROR_SUCCESS;
                    break;
                case NativeMethods.WMIDPREQUESTCODE.WMI_DISABLE_EVENTS:
                    s_traceLog = 0;
                    Status = NativeMethods.ERROR_SUCCESS;
                    break;
                default:
                    Status = NativeMethods.ERROR_INVALID_PARAMETER;
                    break;
            }

            *InOutBufferSize = 0;
            return Status;
        }


        private static uint TagFromString(string strTag)
        {
            uint tag = 0;
            if (!String.IsNullOrEmpty(strTag) && strTag.Length == 4)
            {
                tag = (uint)(strTag[0] << 24 | strTag[1] << 16 | strTag[2] << 8 | strTag[3]);
            }
            return tag;
        }
    }
}
